"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const ts_simple_ast_1 = require("ts-simple-ast");
const ts = __importStar(require("typescript"));
const typescript_ast_util_1 = require("typescript-ast-util");
const typescript_plugins_text_based_user_interaction_1 = require("typescript-plugins-text-based-user-interaction");
const refactors_1 = require("../../refactors");
const reorderParams_1 = require("./reorderParams");
/**
 * # Description
 *
 *  Allows users to change a signature parameter order.
 *
 * # TODO
 *  * refactor a implementation method wont change its interface signature - super
 *  * constructors not supported
 */
class ReorderParamsCodeFixImpl {
    constructor(options) {
        this.options = options;
        this.name = refactors_1.PLUGIN_NAME + '-reorderParams';
        this.config = {
            // print a comment with help on the autocomplete suggestion
            help: false
        };
    }
    description(arg) {
        return this.name;
    }
    apply(arg) {
        const sourceFile = arg.simpleNode.getSourceFile();
        const funcDecl = getSimpleTargetNode(sourceFile, typescript_ast_util_1.positionOrRangeToNumber(arg.positionOrRange), this.selectedAction.args.name, this.options.log);
        if (!funcDecl) {
            this.options.log(`reorderParamsPlugin apply aborted because function ${this.selectedAction.args.name} cannot be found at ${arg.positionOrRange}`);
            return;
        }
        // this.options.log(`reorderParamsPlugin apply ${funcDecl && funcDecl.getKindName()} [${this.selectedAction.args && this.selectedAction.args.reorder && this.selectedAction.args.reorder.join(', ')}] ${funcDecl && funcDecl.getText()}`)
        reorderParams_1.reorderParameters(funcDecl, this.selectedAction.args.reorder, this.options.log);
        sourceFile.saveSync();
        // try {
        //   this.options.info.project.projectService.closeClientFile(arg.sourceFile.fileName)
        //   arg.sourceFile.update(sourceFile.getFullText(), {newLength: sourceFile.getFullText().length,  span: {length:   arg.sourceFile.getFullText().length, start: 0}})
        // }catch(ex){
        //   this.options.log(`reorderParamsPlugin apply ex1: ${ex} ${ex.stack}`)
        // }
    }
    predicate(arg) {
        return true;
    }
    getCompletionsAtPosition(fileName, position, options) {
        return this.getTextUITool().getCompletionsAtPosition(fileName, position, options);
    }
    getApplicableRefactors(info, refactorName, refactorActionName, fileName, positionOrRange, userPreferences) {
        const applicableRefactors = this.getTextUITool().getApplicableRefactors(info, refactorName, refactorActionName, fileName, positionOrRange, userPreferences);
        this.selectedAction = applicableRefactors.selectedAction;
        return applicableRefactors;
    }
    getTextUITool() {
        if (!this.interactionTool) {
            this.interactionTool = typescript_plugins_text_based_user_interaction_1.create(this.getTextUIToolConfig());
        }
        return this.interactionTool;
    }
    getTargetNode(fileName, position) {
        const sourceFile = this.options.info.languageService.getProgram().getSourceFile(fileName);
        const target = typescript_ast_util_1.findChildContainingRangeLight(sourceFile, typescript_ast_util_1.positionOrRangeToRange(position));
        const predicate = p => ts.isCallSignatureDeclaration(p) || ts.isFunctionLike(p) || ts.isCallOrNewExpression(p) || ts.isMethodSignature(p) || ts.isConstructSignatureDeclaration(p);
        this.options.log('getTargetNode  concat(getAscendants, ' + [target].concat(typescript_ast_util_1.getAscendants(target)).map(t => typescript_ast_util_1.getKindName(t) + ' - ' + predicate(t)).join(', '));
        return typescript_ast_util_1.findAscendant(target, predicate, true);
    }
    getTextUIToolConfig() {
        return {
            prefix: '&%&%',
            log: this.options.log,
            actions: [
                {
                    name: 'reorderParams',
                    args: ['name', 'reorder'],
                    commentType: 'block',
                    print: action => `Reorder parameters of "${action.args.name}"`,
                    snippet: (fileName, position) => {
                        let func = this.getTargetNode(fileName, position);
                        if (!func) {
                            return;
                        }
                        let reorder = [];
                        let name;
                        if (ts.isFunctionLike(func)) {
                            if (!func || func.parameters && func.parameters.length <= 1) {
                                return;
                            }
                            name = func.name ? func.name.getText() : (func.parent && func.parent.name && func.parent.name) ? func.parent.name.getText() : undefined;
                            if (!name) {
                                return;
                            }
                            for (let i = 0; i < func.parameters.length; i++) {
                                reorder.push(func.parameters.length - i - 1);
                            }
                        }
                        else if (ts.isCallExpression(func)) {
                            if (!func || func.arguments && func.arguments.length <= 1) {
                                return;
                            }
                            if (!ts.isIdentifier(func.expression)) {
                                return;
                            }
                            name = func.expression.getText();
                            for (let i = 0; i < func.arguments.length; i++) {
                                reorder.push(func.arguments.length - i - 1);
                            }
                        }
                        //TODO: isCallNew and others that comply with : ts.isCallSignatureDeclaration(p)||ts.isFunctionLike(p)||ts.isCallOrNewExpression(p
                        else {
                            this.options.log('snippet undefined because not isCallExpression and not isFunctionLike' + func.getText() + ' - ' + typescript_ast_util_1.getKindName(func));
                            return;
                        }
                        const help = this.config.helpComment ? `
    
            /* Help: The second argument is the new order of parameters. Number N in index M means move the M-th 
                argument to index N. Examples: 
             * [1] means switch between first and second
             * [3, 2] means move the first parameter to fourth position and move the second parameter to the third. 
             *   (the rest of the parameters, (third and fourth) will move left to accommodate this requirements) */` : '';
                        return `reorderParams("${name}", [${reorder.join(', ')}])${help}`;
                    },
                    // TODO: we could give a more intuitive text-based API by letting the user provide the new signature. Then we create a new function with that signature in order to parse it correctly. 
                    nameExtra: (fileName, position) => {
                        const func = this.getTargetNode(fileName, position);
                        if (func && ts.isFunctionLike(func)) {
                            const name = func.name ? func.name.getText() : (func.parent && func.parent.name && func.parent.name) ? func.parent.name.getText() : '';
                            return `of ${name}`;
                        }
                        else if (ts.isCallExpression(func)) {
                            return `of ${func.expression ? func.expression.getText() : ''}`;
                        }
                        else {
                            return '';
                        }
                    }
                }
            ]
        };
    }
}
exports.ReorderParamsCodeFixImpl = ReorderParamsCodeFixImpl;
function getSimpleTargetNode(file, position, name, log) {
    let expr = file.getDescendantAtPos(position);
    if (!expr) {
        return;
    }
    const predicate = e => (ts_simple_ast_1.TypeGuards.isSignaturedDeclaration(e) || ts_simple_ast_1.TypeGuards.isPropertySignature(e) || ts_simple_ast_1.TypeGuards.isFunctionLikeDeclaration(e)) && (ts_simple_ast_1.TypeGuards.isNamedNode(e) || ts_simple_ast_1.TypeGuards.isNameableNode(e) || ts_simple_ast_1.TypeGuards.isPropertyNamedNode(e)) && e.getName() === name;
    // log('getFunctionSimple: '+expr.getKindName() + ' - ' + expr.getText() + ' - ' + expr.getAncestors().map(e=>e.getKindName() + ' - predicate: ' + predicate(e) + ' - pred1: ' + (TypeGuards.isSignaturedDeclaration(e) || TypeGuards.isPropertySignature(e) || TypeGuards.isFunctionLikeDeclaration(e)) + ' - pred2 : ' + (TypeGuards.isNamedNode(e) || TypeGuards.isNameableNode(e) || TypeGuards.isPropertyNamedNode(e))).join(', '));
    const e = [expr].concat(expr.getAncestors()).find(predicate);
    if (!e) {
        return;
    }
    // if (TypeGuards.isSignaturedDeclaration(e) && TypeGuards.isNamedNode(e)) {
    return e;
    // }
}
exports.getSimpleTargetNode = getSimpleTargetNode;
//# sourceMappingURL=reorderParamsPlugin.js.map